#
# MaMa, the Make Master -- a Tcl/Tk script for handling Makefile operations.
#
# Copyright (C) 1998 Asko Seeba.
#
# This file is considered as a part of MaMa package.
#
# The copyright holder disclaims all responsibility or liability with
# respect to its usage or its effect upon hardware or computer
# systems, and maintains copyright as set out in the "LICENCE"
# document which accompanies distributions of MaMa v1.0a0 and upwards.

#######################################################################
# Here comes the declaration of MaMa's initial settings.              #
#######################################################################

wm title . "MaMa v$MAMA_VER"

# Some default values:
# LOC_SIZE     -- width of the column that holds makefile's path and name;
# TARG_SIZE    -- width of the column that holds target's name;
# DESC_SIZE    -- width of the column that holds target's description;
# LETTER_WIDTH -- letter width in pixels in targets listing box.
global LOC_SIZE TARG_SIZE DESC_SIZE LETTER_WIDTH
set LOC_SIZE  21
set TARG_SIZE 7

# Letter size in pixels used in target listing.
set LETTER_WIDTH 7

# Default file of targetlist
set MAMAFILE ""

# Here is the possibility to let a user define his/her own values for variables
# above.
if {[file exists "$RCFILE"]} {
    source $RCFILE
}

source $MAMA_LIB_DIR/help.tcl

#######################################################################
# Here comes the declaration of 'make' driver's procedures.           #
#######################################################################

#
# arrowify <short string> <long string> --
# if <short string> is really shorter than <long string>, replaces the last
# symbol of <short string> with '>'-symbol.
#
proc arrowify {string size} {
    set shortstring [csubstr $string 0 $size]

    if {$size == 0} {
	return ""
    }
    if {$size < [clength $string]} {
	return [cconcat [csubstr $shortstring 0 [expr $size - 1]] >]
    }
    return $shortstring
}

#
# reset_size --
# reconfigure the target list box settings according to ruler and
# window state.
#
proc reset_size {} {
    global LOC_SIZE TARG_SIZE DESC_SIZE LETTER_WIDTH point1Info point2Info

    # update is needed here if window configuration has changed.
    update
    set LOC_SIZE [expr $point1Info(x) / $LETTER_WIDTH - 1]
    set TARG_SIZE [expr ($point2Info(x) - $point1Info(x)) / $LETTER_WIDTH - 1]
    set DESC_SIZE [expr [winfo width .driver.targetbox.list] / \
	    $LETTER_WIDTH - [expr $point2Info(x) / $LETTER_WIDTH] - 1]
}

#
# AddLineToTargetList <line> [index] --
# Add a line <line> to target listing to the location [index]. If
# location isn't specified, add the line to the end of listing.
#
proc AddLineToTargetList {line {index end}} {
    global TargetList

    # Some sanity checks.
    if {[llength $line] == 0} {
	return
    }
    # Remove the whitespaces from beginning of line.
    while {[llength [lindex $line 0]] == 0} {
	set line [lrange $line 1 end]
	if {[llength $line] == 0} {
	    return
    	}
    }

    set TargetList [linsert $TargetList $index $line]

    return $line
}

#
# AddTargetToBox <line> [index] --
# Insert a line <line> to target listing box to the location [index]. If
# location isn't specified, add the line to the end of listing box.
#
proc AddTargetToBox {line {index end}} {
    global TargetList LOC_SIZE TARG_SIZE DESC_SIZE

    # Some sanity checks.
    if {[llength $line] == 0} {
	return
    }
    if {[llength [lindex $line 0]] == 0} {
	return
    }

    # Construct the line
    set line [split $line]
    set line [format \
	    "%-[expr $LOC_SIZE]s %-[expr $TARG_SIZE]s %-[expr $DESC_SIZE]s" \
	    [arrowify [join [lindex $line 0]] $LOC_SIZE] \
	    [arrowify [join [lindex $line 1]] $TARG_SIZE] \
	    [arrowify [join [join [lrange $line 2 end]]] $DESC_SIZE]]
    # Now insert it to the box
    .driver.targetbox.list insert $index $line

    return $line
}

#
# fill_targetbox --
# Fill a target listing box with all targets currently known to MaMa.
#
proc fill_targetbox {} {
    global TargetList

    foreach line $TargetList {
	# If strings are too long, cut them shorter and put an arrow to their
	# end to indicate that actually they are longer. All this will
	# be done in AddTargetToBox.
	AddTargetToBox $line
    }
}

#
# reset_targetheader --
# redraw target list box headers according to its settings.
#
proc reset_targetheader {} {
    global TargetHeader LOC_SIZE TARG_SIZE DESC_SIZE
    set TargetHeader \
      [format "%-[expr $LOC_SIZE]s %-[expr $TARG_SIZE]s %-[expr $DESC_SIZE]s" \
       [arrowify Location    $LOC_SIZE] \
       [arrowify Target      $TARG_SIZE] \
       [arrowify Description $DESC_SIZE]]
}

#
# reset_targetbox --
# redraw target list box according to its settings.
#
proc reset_targetbox {} {
    # Empty the target list at all.
    .driver.targetbox.list delete 0 end

    # Reconfigure the size parameters according to ruler and window size.
    reset_size

    # Redraw target listbox header according to new settings.
    reset_targetheader

    # Fill the target list box with targets again.    
    fill_targetbox
}

#
# load_driver <filename> --
# Load a target listing from file <filename> and display the listing
# in target list box.
#
proc load_driver {filename} {
    global TargetList Save

    set TargetList ""
    set Save 0

    # Sanity check.
    if ![file exists $filename] {
	tk_dialog .filenotfound "File not found" \
		"Can't load file $filename. File not found." error 0 OK
	return
    }
    set filed [open $filename r]
    foreach line [split [read $filed] \n] {
	AddLineToTargetList [split $line]
    }
    close $filed
    reset_targetbox
}

#
# rulerMkTab <ruler> <x> <n> --
# draw an n-th triangle on the ruler with window path name <ruler> at
# the position <x>. <n> >= 1. If <n> is 1, the ruler triangle is
# leftmost.
#
proc rulerMkTab {ruler x n} {
    global point1Info point2Info LETTER_WIDTH

    # Save the information about position.
    set point[expr $n]Info(x) $x

    set y 5
    set size [expr \$point[expr $n]Info(size)]

    $ruler create polygon \
	    [expr $x + 1]         $y \
	    [expr $x + $size - 1] $y \
	    [expr $x + $size / 2] [expr $y + $size]
}

#
# rulerSelectTab <ruler> <x> <n> --
# Select the triangle <n> on ruler, while mouse is pointing on
# cordinate <x> and ruler window path is <ruler>, for further
# operations with that ruler.
#
proc rulerSelectTab {ruler x n} {
    global point1Info point2Info LETTER_WIDTH

    set activeStyle [expr \$point[expr $n]Info(activeStyle)]

    
    set point[expr $n]Info(x) \
	    [expr int([expr [$ruler canvasx $x] / $LETTER_WIDTH]) \
	    * $LETTER_WIDTH]
    $ruler addtag active[expr $n] withtag current
    # That sucks but I have to keep my own track what is selected and what is
    # not because (I don'd understand, why) '$ruler find withtag active1' gives
    # me '1' if the tag is added, but '$ruler find withtag active2' doesn't.
    # That really suck hard -- because of that I have to write more not
    # understandable code.
    set point[expr $n]Info(style) active
    eval "$ruler itemconf active[expr $n] $activeStyle"
    $ruler raise active[expr $n]
}

#
# rulerMoveTab <ruler> <x> --
# Move a triangle on ruler according to mouse x cordinate related to window
# with path name <ruler>. Triangle moves by $LETTER_WIDTH pixels long steps.
#
proc rulerMoveTab {ruler x} {
    global point1Info point2Info LETTER_WIDTH

    # Determine which triangle is moving.
    if {[cequal $point1Info(style) active]} {
	set n 1
    } elseif {[cequal $point2Info(style) active]} {
	set n 2
    } else {
	return
    }

    # If no active triangle doesn't actually exist, return.
    if {[$ruler find withtag active[expr $n]] == ""} {
	return
    }

    # top -- ruler's leftmost top.
    set top  [expr \$point[expr $n]Info(top)]
    set px   [expr \$point[expr $n]Info(x)]

    # Determine the triangle's x cordinate.
    set rulerx [expr int([expr [$ruler canvasx $x] / $LETTER_WIDTH]) \
	    * $LETTER_WIDTH]
    # Determine ruler's width.
    set rulerw [winfo width $ruler]
    set end [expr $rulerw - $top]

    # Check the range in with the triangle is allowed to move. If mouse moves
    # out of the range, don't move the triangle.
    if {$n == 1} {
	# It is leftmost triangle.
	if {$rulerx < $top} {
	    # Mouse moved out of ranges over the ruler's leftmost top.
	    set rulerx $top
	} elseif {[expr $rulerx / $LETTER_WIDTH] >= [expr $point2Info(x) / \
		$LETTER_WIDTH]} {
	    # Mouse moved out of ranges over the second triangle.
	    set rulerx [expr $point2Info(x) - $LETTER_WIDTH]
	}
    } elseif {$n == 2} {
	# It is second triangle
	if {[expr $rulerx / $LETTER_WIDTH] <= \
		[expr $point1Info(x) / $LETTER_WIDTH]} {
	    # Mouse moved out of ranges over the leftmost triangle.
	    set rulerx [expr $point1Info(x) + $LETTER_WIDTH]
	} elseif {$rulerx > $end} {
	    # Mouse moved out of ranges over the ruler's rightmost top.
	    set rulerx $end
	}
    }
    # Place the triangle to the corresponding location.
    $ruler move active[expr $n] [expr $rulerx - $px] 0
    # Remember the triangle's x cordinate.
    set point[expr $n]Info(x) $rulerx
}

#
# rulerReleaseTab <ruler> <x> --
# deactivate the triangle in the ruler window <ruler> at cordinate <x>.
#
proc rulerReleaseTab {ruler x} {
    global point1Info point2Info LETTER_WIDTH LOC_SIZE TARG_SIZE

    # Determine which triangle we are dealing with.
    if {[cequal $point1Info(style) active]} {
	set n 1
    } else {
	set n 2
    }
    set normalStyle [expr \$point[expr $n]Info(normalStyle)]

    # If the triangle wasn'r activated actually, we haven't anything to do
    # more.
    if {[$ruler find withtag active[expr $n]] == {}} {
	return
    }
    # Deactivate the triangle.
    eval "$ruler itemconf active[expr $n] $normalStyle"
    # Remeber the triangle's status.
    set point[expr $n]Info(style) normal
    # Delete the tag that implements triangle's activated status.
    $ruler dtag active[expr $n]

    reset_targetbox
}

#
# init_ruler <ruler> --
# unitialize the ruler with window path name <ruler>
#
proc init_ruler {ruler} {
    global point1Info point2Info LETTER_WIDTH LOC_SIZE TARG_SIZE

    # Each triangle on ruler has it's configuration information.
    set point1Info(grid) $LETTER_WIDTH
    set point2Info(grid) $LETTER_WIDTH
    set point1Info(top)  [expr 5 * $LETTER_WIDTH]
    set point2Info(top)  [expr 5 * $LETTER_WIDTH]
    set point1Info(size) $LETTER_WIDTH
    set point2Info(size) $LETTER_WIDTH
    set point1Info(normalStyle) "-fill black"
    set point2Info(normalStyle) "-fill black"
    if {[winfo depth $ruler] > 1} {               # We have some colours
	set point1Info(activeStyle) "-fill red -stipple {}"
	set point2Info(activeStyle) "-fill red -stipple {}"
    } else {
	set point1Info(activeStyle) "-fill black -stipple {}"
	set point2Info(activeStyle) "-fill black -stipple {}"
    }

    # Determine the location of first triangle and place the triangle to the
    # location.
    set x [expr ($LOC_SIZE + 1) * $LETTER_WIDTH]
    $ruler addtag tab1 withtag \
	    [rulerMkTab $ruler [expr ($LOC_SIZE + 1) * $LETTER_WIDTH] 1]
    set point1Info(style) active
    rulerMoveTab $ruler $x
    set point1Info(style) normal

    # Determine the location of second triangle and place the triangle to the
    # location.
    set x [expr ($LOC_SIZE + $TARG_SIZE + 2) * $LETTER_WIDTH]
    $ruler addtag tab2 withtag \
	    [rulerMkTab \
	     $ruler [expr ($LOC_SIZE + $TARG_SIZE + 2) * $LETTER_WIDTH] 2]
    set point2Info(style) active
    rulerMoveTab $ruler $x
    set point2Info(style) normal

    # Set the bindings to ruler.
    $ruler bind tab1 <1> "rulerSelectTab $ruler %x 1"
    $ruler bind tab2 <1> "rulerSelectTab $ruler %x 2"
    bind $ruler <B1-Motion>           "rulerMoveTab $ruler %x"
    bind $ruler <Any-ButtonRelease-1> "rulerReleaseTab $ruler %x"
}

#
# reset_output --
# clear the make execution's output box.
#
proc reset_output {} {
    .output.text configure -state normal
    .output.text delete 1.0 end
    .output.text configure -state disabled
}

#
# DoMake --
# execute a 'make' program according to selected element in target list box.
#
proc DoMake {} {
    global TargetList
    global MakingAlready

    if {$MakingAlready == 1} {
	tk_dialog .makingalready "Making Already" \
		"There is an unfinished making job already. Possibly You gave another command accidentally. I'll make the unfinished one only." error 0 OK
	return
    }
    set MakingAlready 1
    # Sanity check
    set curselection [.driver.targetbox.list curselection]
    if {[llength $curselection] == 0} {
	NoSelection make
	set MakingAlready 0
	return
    }
    # Determine the location of makefile.
    set location [lindex [lindex $TargetList \
	    [.driver.targetbox.list curselection]] 0]

    # Sanity check if makefile exists and is accessible.
    if {[file exists $location] == 0} {
	tk_dialog .dirnotfound "Error in targets list" \
		"File $location not found." error 0 OK
	set MakingAlready 0
	return
    }

    # Determine the path and makefilename from location string.
    # $dir will be path to makefile and $makefile will be makefile's
    # name.
    set splited_loc [split $location /]
    set path_length [llength $splited_loc]
    if {[cindex $location 0] == "/"} {
	set dir "/"
    } else {
	set dir         ""
    }
    foreach step [lrange $splited_loc 0 [expr $path_length - 2]] {
	if {[clength $dir] != 0} {
	    set dir $dir/$step
	} else {
	    set dir $step
	}
    }
    set makefile [lindex $splited_loc end]

    # Determine the target name.
    set target [join [lindex [split [lindex $TargetList \
	    [.driver.targetbox.list curselection]]] 1]]

    # Go to the location of makefile if specified.
    if {[clength $dir] > 0} {
        cd $dir
    }
    # Clear the make execution's output window.
    reset_output


    # This will be the command executed.
    # NB! There MUST be a SPACE BETWEEN -f and $makefile!!!
    #     Otherwise some make programs generate an error.
    set make_command "make"
    if {[string length $makefile] > 0} {
	# The situation that makefile isn't specified shouldn't occur, while
	# makefile name must be specified in target listing, but... who
	# knows...
	set make_command "$make_command -f $makefile"
    }
    if {[string length $target]} {
	set make_command "$make_command $target"
    }

    # '|& cat' is for redirecting standard error to standard out.
    set piped [open "|[split $make_command] |& cat" r]

    # I want to show the generated output by 'make' right after output is
    # generated by 'make'. I don't want that the output appears only after
    # 'make' has finished. So, set the pipe to nonblocking mode.
    fconfigure $piped -blocking false

    # Read from the pipe until EOF will be encountered i.e. until 'make' has
    # finished (and closed its side of the pipe).
    while {![eof $piped]} {
	# Wait for something appears in pipe.
	if {[llength [select $piped {} {} 5]] == 0} {
	    # If pipe was empty, do nothing.
	    continue
	} else {
	    # Allow to write to the output box, write the contents of the
	    # pipe to the output box, disallow the modifications of the
	    # output box again, show the end of the output box and force
	    # wishx to show the changes made on display.
	    .output.text configure -state normal
	    .output.text insert end "[read $piped]"
	    .output.text configure -state disabled
	    .output.text see end
	    update
	}
    }
    set retval [catch "close $piped"]

    # 'make' has finished. Show the informative message about that and the
    # end of the output box.
    .output.text configure -state normal
    .output.text insert end \
	    "\nMaMa: '$make_command' completed with exit status $retval."
    .output.text configure -state disabled
    .output.text see end
    set MakingAlready 0
    return
}

#
# positionWindow <w> [x] [y] --
# set the toplevel window's geometry's x-cordinate to <x> and y-cordinate to
# <y>. Defaults are 300 and 200 accordingly.
#
proc positionWindow {w {x 300} {y 200}} {
    wm geometry $w +$x+$y
}

#
# MakefileDialog <parent> <entry> --
# display a file browser dialog box with parent window path name <parent> for
# specifying makefile name and place the browser's result to the entry window
# with path name <entry>.
#
proc MakefileDialog {parent entry} {
    #   Type names		Extension(s)
    #
    #---------------------------------------
    set types {
	{"All files"		*}
    }

    set file [tk_getOpenFile -filetypes $types -parent $parent \
	    -defaultextension *]

    if ![cequal $file ""] {
	$entry delete 0 end
	$entry insert 0 $file
	$entry xview end
    }
}

#
# AddTarget <line> <index> --
# add a new target description to known target descriptions list and target's
# list box.
#
proc AddTarget {line index} {
    AddLineToTargetList $line $index
    AddTargetToBox      $line $index
}

#
# Terminate <w> <caller> [index] --
# terminate the new target addition from target editing dialog box'es, with
# window path name <w> contents. <caller> is the buttons window path name
# according to the command that caused the displaying of target edition dialog
# box and index is the location of the target edited in the dialog box.
#
proc Terminate {w caller {index end}} {
    global Save
    set f $w.data

    # Some sanity checks.
    set makefile [$f.makefile get]
    if {[clength $makefile] == 0} {
	tk_dialog $w.error "No makefile" \
		"Makefile must be specified." error 0 OK
	return
    }
    if {![file exists $makefile]} {
	tk_dialog $w.error "Illegal makefile" \
		"Makefile $makefile doesn't exist." error 0 OK
	return
    }

    # Construct the line according to the dialog box'es contents.
    set line [split [$f.makefile get]]
    set line "$line [split [$f.target get]]"
    set line "$line [split [$f.comment get]]"

    # Add a target to known targets list and to the target list box.
    AddTarget $line $index

    # List's contents has changed now. Ask a saving verification if needed.
    set Save 1

    # Destroy the target editing dialog box.
    DestroyTED $w $caller
}

#
# DestroyTED <w> <caller> --
# destroy a target edit dialog box with window path name <w> and put the focus
# on button according to the command that displayed the dialog box.
#
proc DestroyTED {w caller} {
    focus $caller
    destroy $w
}

#
# TargetEditDialog <w> <title> <iconname> <terminate> <caller> --
# display the target editing dialog box with window path name <w>, window
# title name <title>, window's iconname <iconname>, command <terminate> binded
# to 'OK' button on that window. <caller> is the button's window path name
# according to the command that uses this procedure here.
#
proc TargetEditDialog {w title iconname terminate caller} {
    # Construct a toplevel window.
    catch {destroy $w}
    toplevel $w
    wm title $w $title
    wm iconname $w $iconname
    positionWindow $w

    # Build up a layout of the dialog box.
    frame  $w.data -relief ridge -borderwidth 3
    pack   $w.data -side top
    
    label  $w.data.makefileprompt -text Makefile:
    grid   $w.data.makefileprompt -row 0 -column 0 -sticky e
    entry  $w.data.makefile
    grid   $w.data.makefile       -row 0 -column 1 -sticky w
    button $w.data.filebrowse     -text Browse \
	    -command "MakefileDialog $w.data $w.data.makefile"
    grid   $w.data.filebrowse     -row 0 -column 2

    label  $w.data.targetprompt   -text Target:
    grid   $w.data.targetprompt   -row 1 -column 0 -sticky e
    entry  $w.data.target
    grid   $w.data.target         -row 1 -column 1 -sticky w

    label  $w.data.commentprompt  -text Comment:
    grid   $w.data.commentprompt  -row 2 -column 0 -sticky e
    entry  $w.data.comment
    grid   $w.data.comment        -row 2 -column 1 -sticky w

    frame  $w.terminate -relief ridge -borderwidth 3
    pack   $w.terminate -side top -fill x
    
    button $w.terminate.ok        -text OK -command $terminate
    grid   $w.terminate.ok        -row 0 -column 0 -padx 30

    button $w.terminate.cancel    -text Cancel -relief raised \
	    -command "DestroyTED $w $caller"
    grid   $w.terminate.cancel    -row 0 -column 1 -padx 30

    focus $w.data.makefile

    # Set the bindings.
    bindtags $w
    # {all . Listbox .driver.targetbox.list}
    bind $w                  <Escape> "DestroyTED $w $caller"
    bind $w.data.comment     <Return> $terminate
    bind $w.data.filebrowse  <Return> "MakefileDialog $w.data $w.data.makefile"
    bind $w.data.makefile    <Return> $terminate
    bind $w.data.target      <Return> $terminate
    bind $w.terminate.ok     <Return> $terminate
    bind $w.terminate.cancel <Return> "DestroyTED $w $caller"
}

#
# TerminateNewTarget <w> <caller> <index> --
# terminate the addition of new target to the location <index> in target list,
# edited in edit dialog box with window path name <w>. <caller> is the button's
# window path name according to the command that displayed the target editing
# dialog box.
#
proc TerminateNewTarget {w caller index} {
    Terminate $w $caller $index
}

#
# NewTarget <w> <index> <verb> <caller> --
# display the target editing dialog box with window path name <w> for adding
# new target to location <index> to the target list. <verb> is a word used for
# constructin a string for target editing dialog box'es title and icon name.
# <caller> is a button's window path name according to the command that
# displayed the target editing dialog box. If 'Makefile' exists in current
# directory, it will be inserted to the makefile name field as a default value.
#
proc NewTarget {w index verb caller} {
    TargetEditDialog $w "$verb new target" $verb \
	    "TerminateNewTarget $w $caller $index" $caller
    if {[file exists Makefile] == 1} {
	$w.data.makefile insert end Makefile
    } elseif {[file exists makefile] == 1} {
	$w.data.makefile insert end makefile
    }
}

#
# DoAdd --
# add a target after selected element into target list. If no element is
# selected, append the target to the end of list.
#
proc DoAdd {} {
    set index [.driver.targetbox.list curselection]
    if {$index == "" } {
	set index end
    } else {
	set index [expr $index + 1]
    }
    NewTarget .add $index Add .driver.commands.edit.add
}

#
# DoInsert --
# insert a target before selected element into target list. If no element is
# selected, prepend the target to the beginning of list.
#
proc DoInsert {} {
    set index [.driver.targetbox.list curselection]
    if {$index == "" } {
	set index 0
    }
    NewTarget .insert $index Insert .driver.commands.edit.insert
}

#
# DeleteTarget <index> --
# delete the target from list and listing box at the index <index>.
#
proc DeleteTarget {index} {
    global TargetList
    set TargetList [lreplace $TargetList $index $index]
    .driver.targetbox.list delete $index $index
}

#
# ClearTargetList --
# empty the target list and listing box.
#
proc ClearTargetList {} {
    global TargetList

    # If list is empty already, do nothing.
    if {[llength $TargetList] == 0} {
	return
    }

    set TargetList [lreplace $TargetList 0 end]
    .driver.targetbox.list delete 0 end
}

#
# TerminateEdit <caller> <index> --
# delete an old entry from target list and insert the target with new settings
# specified in target edit dialog box to the same location to target list.
# In other words -- change the target's settings in target list.
#
proc TerminateEdit {caller index} {
    DeleteTarget $index
    Terminate .edit $caller $index
}

#
# NoSelection <do_something> --
# display an error message "You have to select something to <do_something>".
#
proc NoSelection {do_something} {
    tk_dialog .no_selection "Nothing selected" \
	    "You have to select something to $do_something." error 0 OK
}

#
# DoEdit --
# edit a selected target in target list. If no target selected, display an
# according error message.
#
proc DoEdit {} {
    global TargetList

    # Sanity check.
    set index [.driver.targetbox.list curselection]
    if {$index == "" } {
	NoSelection edit
	return
    }

    # Determine the according line in target list.
    set line  [lindex $TargetList $index]

    # Display target edit dialog box and fill its field with according values
    # from line from target listing.
    TargetEditDialog .edit "Edit a target" Edit \
	    "TerminateEdit .driver.commands.edit.edit $index" \
	    .driver.commands.edit.edit
    .edit.data.makefile insert end [lindex $line 0]
    .edit.data.target   insert end [lindex $line 1]
    .edit.data.comment  insert end [join [lrange $line 2 end]]
}

#
# Verify <do_something> [with_it] <index> --
# ask a verification for <do_something> [with_it]. <index> is a default answer
# 0 or 1 and that means 'Yes' or 'No' accordingly.
#
proc Verify {do_something {with_it ""} index} {
    return [tk_dialog .verify "Verify" \
	    "Are You sure you want to $do_something $with_it?" \
	    question $index Yes No]
}

#
# DoDelete --
# delete a selected target from target list. If no target selected, display an
# according error message.
#
proc DoDelete {} {
    global Save

    set index [.driver.targetbox.list curselection]
    if {$index != ""} {
	if {[Verify delete "selected target" 1] == 0} {
	    DeleteTarget $index
	    set Save 1
	}
    } else {
	NoSelection delete
    }
}

#
# copy <filename1> <filename2> --
# copy a file <filename1> to file <filename2>.
#
proc copy {filename1 filename2} {
    set from [open $filename1 r]
    set to   [open $filename2 w]
    copyfile $from $to
    close $from
    close $to
}

#
# save <filename> --
# save target list to file <filename>.
#
proc save {filename} {
    global TargetList

    set fileId [open $filename w]
    foreach line $TargetList {
	set l [split $line]
	set location    [join [lindex $l 0]]
	set target      [join [lindex $l 1]]
	set description [join [join [lrange $l 2 end]]]
	puts $fileId "$location $target $description"
    }
    close $fileId
}

#
# DoNew --
# create a new target list. If there are unsaved changes in current target
# list, ask for verification and give a possibility to save the changes.
#
proc DoNew {} {
    global Save FileName

    if {$Save == 1} {
	switch [AskForSave "starting a new targetlist"] \
		0 DoSave \
		1 {} \
		default return
    }

    ClearTargetList
    set FileName ""
    set Save 0
}

#
# SelectFileName --
# display a file selection browser for selecting a file for saving target list.
#
proc SelectFileName {} {
    #   Type names		Extension(s)
    #
    #---------------------------------------
    set types {
	{"MaMa files"           .mam}
	{"All files"		*}
    }
    return [tk_getSaveFile -filetypes $types -parent .driver.commands \
	    -defaultextension .mam]
}

#
# DoSave --
# save a target list. If filename isn't known, ask for filename too. Copy an
# existing file with the same name to file with the same name with suffix .bak.
# RETURN: 1 if saved, 0 if not saved.
#
proc DoSave {} {
    global FileName Save

    if {$FileName == ""} {
	set file [SelectFileName]
	if [cequal $file ""] {
	    tk_dialog .nosavefilename "No file name" \
		    "No file specified for saving. Targetlist not saved." \
		    info 0 OK
	    return 0
	}
	set FileName $file
    }

    if {[file exists $FileName]} {
	copy $FileName $FileName.bak
    }
    save $FileName
    tk_dialog .savecompleted "File saved" \
	    "File $FileName saved." info 0 OK
    set Save 0
    return 1
}

#
# DoSaveAs --
# ask a file name to save a target list and save the target list to file with
# the specified filename.
#
proc DoSaveAs {} {
    global FileName

    # Actually run a DoSave procedure, but before that forget the old file
    # name.
    set FileName ""
    DoSave
}

#
# AskForSave <action> --
# give a possibility to save target list before <action>.
#
proc AskForSave {action} {
    return [tk_dialog .unsavedchanges "Unsaved changes" \
	    "You have unsaved changes in make list. Save before $action?" \
	    question 2 Yes No Cancel]
}

#
# DoExit --
# exit MaMa. Also give a possibility to save unsaved changes in target list.
#
proc DoExit {} {
    global Save

    if {[tk_dialog .unsavedchanges "Verify exit" \
	    "Do You really want to exit MaMa?" question 1 Yes No] == 1} {
	return
    }
    if {$Save == 1} {
	switch [AskForSave exit] \
		0       {if {[DoSave] == 1} {destroy .} else {return}} \
		1       {destroy .} \
		default return
    }
    destroy .
}

#
# DoOpen --
# ask a file name and load a target list from the specified file.
#
proc DoOpen {} {
    global FileName Save

    if {$Save == 1} {
	switch [AskForSave "loading new make list"] \
		0       DoSave \
		1       {} \
		default return
    }

    #   Type names		Extension(s)
    #
    #---------------------------------------
    set types {
	{"MaMa files"           .mam}
	{"All files"		*}
    }
    # File browser dialog box.
    set file [tk_getOpenFile -filetypes $types -parent .driver.commands \
	    -defaultextension .mam]
    if ![cequal $file ""] {
	set FileName $file
	load_driver $FileName
    }
}

#
# GetTarget <index> --
# determine the target at the location <index> in target list.
#
proc GetTarget {index} {
    global TargetList

    return [lindex $TargetList $index]
}

#
# SelectTarget <index> --
# implement the selection of target at the location <index> in target list.
#
proc SelectTarget {index} {
    .driver.targetbox.list selection set $index
    .driver.targetbox.list activate $index
}

#
# DoUp --
# move the selected line in target listing box one step up. If line is already
# at the top of listing, do nothing. If ther is no line selected, display an
# error message.
#
proc DoUp {} {
    global Save

    set index [.driver.targetbox.list curselection]
    # Sanity checks.
    if {$index == "" } {
	NoSelection "move it up"
	return
    }

    if {$index == 0} {
	return
    }

    # The selected line has to move.
    set Save 1
    set target [GetTarget $index]
    DeleteTarget $index
    set newindex [expr $index - 1]
    AddTarget $target $newindex
    # Because of there is also a default binding of up-arrow key in listbox --
    # move the selection up, we have to select the line at old index. After
    # this procedure the default binding takes place and the selection will be
    # then on just the line That has moved up.
    SelectTarget $index
}

#
# DoDown --
# move the selected line in target listing box one step down. If line is
# already at the bottom of listing, do nothing. If ther is no line selected,
# display an error message.
#
proc DoDown {} {
    global Save

    set index [.driver.targetbox.list curselection]
    if {$index == "" } {
	NoSelection "move it up"
	return
    }

    if {$index == [expr [.driver.targetbox.list size] - 1]} {
	return
    }

    set Save 1
    set target [GetTarget $index]
    DeleteTarget $index
    set newindex [expr $index + 1]
    AddTarget $target $newindex
    # Because of there is also a default binding of down-arrow key in listbox
    # -- move the selection down, we have to select the line at old index.
    # After this procedure the default binding takes place and the selection
    # will be on just the line that has moved down.
    SelectTarget $index
}

#
# not_implemented_yet <detail> --
# display a message "<detail> not implemented yet.".
#
proc not_implemented_yet {detail} {
    tk_dialog .notimplementedyet "Not implemented" \
	    "$detail not implemented yet." info 0 OK
}

#######################################################################
# The declaration of procedures is over.                              #
#######################################################################
# Here comes the declaration of MaMa's main window's layout.          #
#######################################################################

# Create a main menu frame.
frame .mainmenu -relief raised -borderwidth 1
pack  .mainmenu -side top -fill x

wm iconname . MaMa
wm minsize  . 500 500
wm geometry . 550x500+5+5

#
# Fill a main menu with buttons.
#

# Width of menubuttons.
set MBWIDTH 7

# A file menu.
menubutton .mainmenu.file -text File -underline 0 -menu .mainmenu.file.m \
	-relief raised -width $MBWIDTH
pack       .mainmenu.file -side left
menu       .mainmenu.file.m
.mainmenu.file.m add command -label New           -accelerator (Ctrl+n) \
	-command DoNew
.mainmenu.file.m add command -label Open          -accelerator (Ctrl+o) \
	-command DoOpen
.mainmenu.file.m add command -label Save          -accelerator (Ctrl+s) \
	-command DoSave
.mainmenu.file.m add command -label "Save as ..." -accelerator (Ctrl+S) \
	-command DoSaveAs
.mainmenu.file.m add separator
.mainmenu.file.m add command -label "Exit MaMa"   -accelerator (Esc) \
	-command DoExit

# An edit menu.
menubutton .mainmenu.edit -text Edit -underline 0 -menu .mainmenu.edit.m \
	-relief raised -width $MBWIDTH
pack .mainmenu.edit -side left
menu .mainmenu.edit.m
.mainmenu.edit.m add command -label Insert -accelerator (Ctrl+i) \
	-command DoInsert
.mainmenu.edit.m add command -label Add    -accelerator (Ctrl+a) \
	-command DoAdd
.mainmenu.edit.m add command -label Edit   -accelerator (Ctrl+e) \
	-command DoEdit
.mainmenu.edit.m add command -label Up     -accelerator (Ctrl+<Up>) \
	-command DoUp
.mainmenu.edit.m add command -label Down   -accelerator (Ctrl+<Down>) \
	-command DoDown

# An options menu.
menubutton .mainmenu.options -text Options -underline 0 \
	-menu .mainmenu.options.m -relief raised -width $MBWIDTH
pack       .mainmenu.options -side left
menu       .mainmenu.options.m
.mainmenu.options.m add command -label "Not yet implemented..." -state disabled
.mainmenu.options.m add separator
.mainmenu.options.m add command -label "Here will come"         -state disabled
.mainmenu.options.m add command -label "configuration commands" -state disabled
.mainmenu.options.m add command -label "for MaMa in future."    -state disabled

# A run menu.
menubutton .mainmenu.run -text Run -underline 0 -menu .mainmenu.run.m \
	-relief raised -width $MBWIDTH
pack       .mainmenu.run -side left
menu       .mainmenu.run.m
.mainmenu.run.m add  command -label Make -accelerator (<Return>) \
	-command DoMake

# A help menu.
menubutton .mainmenu.help -text Help -underline 0 -menu .mainmenu.help.m \
	-relief raised -width $MBWIDTH
# flat groove raised ridge sunken
pack       .mainmenu.help -side right
menu       .mainmenu.help.m
.mainmenu.help.m add command -label About -accelerator (Ctrl+A) \
	-command ShowAbout
.mainmenu.help.m add command -label Introduction -accelerator (Ctrl+I) \
	-command ShowIntroduction
.mainmenu.help.m add command -label "Getting Started" -accelerator (Ctrl+G) \
	-command ShowGettingStarted
.mainmenu.help.m add command -label Tutorial -accelerator (Ctrl+R) \
	-command ShowTutorial
.mainmenu.help.m add command -label Controls -accelerator (Ctrl+C) \
	-command ShowControls
.mainmenu.help.m add command -label Done -accelerator (Ctrl+D) \
	-command ShowDone
.mainmenu.help.m add command -label Ideas -accelerator (Ctrl+E) \
	-command ShowIdeas
.mainmenu.help.m add command -label "Known Problems" -accelerator (Ctrl+K) \
	-command ShowKnownProblems
.mainmenu.help.m add command -label Support -accelerator (Ctrl+P) \
	-command ShowSupport

#
# Construct the layout of driver for 'make' program.
#

frame .driver -relief ridge -borderwidth 3
pack  .driver -fill both -expand yes

frame     .driver.targetbox -relief ridge -borderwidth 3
pack      .driver.targetbox -fill both -expand yes

# Ruler for width of columns in targets listing box.
if {[winfo depth .driver.targetbox] > 1} {       # We have some colours
    canvas    .driver.targetbox.ruler -height 10 -relief ridge
} else {
    canvas    .driver.targetbox.ruler -height 10 -relief ridge \
	    -background white
}

grid      .driver.targetbox.ruler -row 0 -column 0 -sticky ew

# The header of the columns in listbox.
global TargetHeader
label     .driver.targetbox.header -textvar TargetHeader \
	-font "[expr $LETTER_WIDTH]x14" -anchor w
grid      .driver.targetbox.header -row 1 -column 0 -sticky w

listbox   .driver.targetbox.list -relief sunken \
	-font "[expr $LETTER_WIDTH]x14" \
	-yscroll ".driver.targetbox.yscroll set"
grid      .driver.targetbox.list -row 2 -column 0 -sticky news
# The listbox should be the onlyone who expands to somewhere.
grid rowconfig    .driver.targetbox 2 -weight 1
grid columnconfig .driver.targetbox 0 -weight 1

scrollbar .driver.targetbox.yscroll \
	-command ".driver.targetbox.list yview" -orient vertical
grid      .driver.targetbox.yscroll -row 2 -column 1 -sticky ns

#
# Here comes the button table at the left side of MaMa.
#

frame .driver.commands -relief ridge -borderwidth 3
# We have to place the buttons frame on left side before listbox on right side
pack  .driver.commands -side right -fill y -before .driver.targetbox

button .driver.commands.make -text Make   -command DoMake
pack   .driver.commands.make -side top -padx 10 -fill x

frame  .driver.commands.edit -relief ridge -borderwidth 3
pack   .driver.commands.edit -side top -fill x
button .driver.commands.edit.insert -text Insert -command DoInsert
button .driver.commands.edit.add    -text Add    -command DoAdd
button .driver.commands.edit.edit   -text Edit   -command DoEdit
button .driver.commands.edit.delete -text Delete -command DoDelete
button .driver.commands.edit.up   -text Up   -command DoUp
button .driver.commands.edit.down -text Down -command DoDown
grid   .driver.commands.edit.insert -row 0 -column 0 -sticky new
grid   .driver.commands.edit.add    -row 0 -column 1 -sticky new
grid   .driver.commands.edit.edit   -row 1 -column 0 -sticky new
grid   .driver.commands.edit.delete -row 1 -column 1 -sticky new
grid   .driver.commands.edit.up     -row 2 -column 0 -sticky new
grid   .driver.commands.edit.down   -row 2 -column 1 -sticky new

frame  .driver.commands.file -relief ridge -borderwidth 3
pack   .driver.commands.file -side top -fill x
button .driver.commands.file.new    -text New       -command DoNew
button .driver.commands.file.open   -text Open      -command DoOpen
button .driver.commands.file.save   -text Save      -command DoSave
button .driver.commands.file.saveas -text "Save as" -command DoSaveAs
grid   .driver.commands.file.new    -row 0 -column 0 -sticky new
grid   .driver.commands.file.open   -row 0 -column 1 -sticky new
grid   .driver.commands.file.save   -row 1 -column 0 -sticky new
grid   .driver.commands.file.saveas -row 1 -column 1 -sticky new

button .driver.commands.exit -text Exit -command DoExit
pack   .driver.commands.exit -side bottom -padx 10 -fill x

#
# Construct the layout for output of executed system commands with usefull.
# output.
#

frame     .output -relief ridge -borderwidth 3
pack      .output -side top -fill both -expand yes

text      .output.text -relief sunken -state disabled \
	-font "[expr $LETTER_WIDTH]x14" \
	-yscroll ".output.yscroll set"
grid      .output.text -row 0 -column 0 -sticky news
# The listbox should be the onlyone who expands to somewhere.
grid rowconfig    .output 0 -weight 1
grid columnconfig .output 0 -weight 1

scrollbar .output.yscroll -command ".output.text yview" \
	-orient vertical
grid      .output.yscroll -row 0 -column 1 -sticky ns

#
# Proceed some more initialization work provided by procedures.
#

init_ruler .driver.targetbox.ruler

global FileName
if ![cequal $MAMAFILE ""] {
    set FileName $MAMAFILE
    load_driver $FileName
} else {
    global TargetList Save

    set TargetList {}
    set FileName ""
    set Save 0
}

global MakingAlready
set MakingAlready 0

#
# Bind some events.
#

# Following two lines avoid a default possibility to change the xview of
# listbox with B2-Motion. The horizontal virtual size of that listbox
# appears wider than actual size, if some line has a '>' sign at the end of
# "Comment" field.
bind .driver.targetbox.list <B2-Motion>   {.driver.targetbox.list xview 0}
bindtags .driver.targetbox.list {all . Listbox .driver.targetbox.list}

bind .                            <Control-a>            DoAdd
bind .                            <Control-d>            DoDelete
bind .                            <Control-Down>         DoDown
bind .                            <Control-e>            DoEdit
bind .                            <Control-c>            DoExit
bind .                            <Escape>               DoExit
bind .                            <Control-i>            DoInsert
bind .                            <Control-n>            DoNew
bind .                            <Control-o>            DoOpen
bind .                            <Control-s>            DoSave
bind .                            <Control-S>            DoSaveAs
bind .                            <Control-Up>           DoUp
bind .                            <Control-A>            ShowAbout
bind .                            <Control-C>            ShowControls
bind .                            <Control-D>            ShowDone
bind .                            <Control-E>            ShowIdeas
bind .                            <Control-G>            ShowGettingStarted
bind .                            <Control-I>            ShowIntroduction
bind .                            <Control-K>            ShowKnownProblems
bind .                            <Control-R>            ShowTutorial
bind .                            <Control-P>            ShowSupport
bind .driver.commands.exit        <Return>               DoExit
bind .driver.commands.edit.add    <Return>               DoAdd
bind .driver.commands.edit.insert <Return>               DoInsert
bind .driver.commands.edit.delete <Return>               DoDelete
bind .driver.commands.edit.edit   <Return>               DoEdit
bind .driver.commands.file.new    <Return>               DoNew
bind .driver.commands.file.open   <Return>               DoOpen
bind .driver.commands.file.save   <Return>               DoSave
bind .driver.commands.file.saveas <Return>               DoSaveAs
bind .driver.commands.make        <Return>               DoMake
bind .driver.targetbox.list       <ButtonPress> \
	{focus .driver.targetbox.list}
bind .driver.targetbox.list       <Double-ButtonPress-1> DoMake
bind .driver.targetbox.list       <Configure>            reset_targetbox
bind .driver.targetbox.list       <Return>               DoMake

#
# Have to define an action to SIGCHLD because of executing make program using
# pipe.
#
#signal ignore SIGCHLD

#
# Let's rock now!
#

